// SerialFunctions.h
/* Sam Bobb,  Daniel Cornew,  Ryan Deeter
	Two degree of freedom arm
	High speed motor control for 2 motors
	Code version 3
	March 18, 2010
	

*/

// Note: To malloc anything, you have to set a heap size in the linker options.
// In MPLAB:  Project -> Build Options -> Project
// Go to tab MPLAB PIC32 Linker
// Set heap size to something big enough to hold the largest ammount of mem you might malloc

#ifndef __SERIALFUNCTIONS_H
#define __SERIALFUNCTIONS_H

#define SHORT_INT_BYTELENGTH		2

#define TIMEOUT_COUNT 50000

//This union is used to split a 16-bit signed int into 2 bytes	
// (Remember your C: the elements of a union occupy the same memory.
//  So in this case, you can access that memory as either a short int or as a two element char array)
union serialShortInt{
	signed short int dataShortInt;
	unsigned char dataChar[SHORT_INT_BYTELENGTH];
};

/** Function Declarations **************************************/

void process_incoming_data();

signed short int read_short_int();

void wait_timeout();

void sendDataRS232(void);

/** Functions **************************************/

void process_incoming_data()
{
	char data;
	int i = 0;
	short int numPathPoints;
	
	switch(read_short_int())
	{		
		case C_HOLD:
			// stop running control
			holdArm = FALSE;
			runcontrol = FALSE;
			// set the hold point
			holdTheta[0] = read_short_int();
			holdTheta[1] = read_short_int();
			// clear everything
			for (i = 0; i<ERROR_DERIV_DELAY; i++) 
			{
				errorBuffer[0][i] = 0;
				errorBuffer[1][i] = 0;
			}
			errorIndex = 0;
			globalIndex = 0;
			controlCount = 0; 
			// run control again
			holdArm = TRUE;
			runcontrol = FALSE;
			break;
		
		case C_FLOAT:
			// Stop and float motors
			runcontrol2 = read_short_int();
			if(runcontrol2 == 1){
				holdArm = FALSE;
				motor_float();
			}
			break;
			
		case C_RUN:
			// Run
			runcontrol2 = read_short_int();
			if(runcontrol2 == 1){
				// stop running control
				holdArm = FALSE;
				runcontrol = FALSE;
				// set our hold point
				holdTheta[0] = currentTheta[0];
				holdTheta[1] = currentTheta[1];
				// clear errors
				for (i = 0; i<ERROR_DERIV_DELAY; i++) 
				{
					errorBuffer[0][i] = 0;
					errorBuffer[1][i] = 0;
				}
				errorIndex = 0;
				// start motors
				motor_run();
			}
			break;
			
		case C_CONTROL_CONFIG:
			// Set new control params
			KPnum[0] = read_short_int();
			KPden[0] = read_short_int();
			KDnum[0] = read_short_int();
			KDden[0] = read_short_int();
			KInum[0] = read_short_int();
			KIden[0] = read_short_int();
			Kslope[0] = read_short_int();
			
			KPnum[1] = read_short_int();
			KPden[1] = read_short_int();
			KDnum[1] = read_short_int();
			KDden[1] = read_short_int();
			KInum[1] = read_short_int();
			KIden[1] = read_short_int();
			Kslope[1] = read_short_int();
			break;
			
		case C_LOG_CONFIG:
			// Send new logging params
			// not doing anything here yet
			break;
		
		case C_SET_PATH:
			// Pause, Create path, hold at current
			// Stop running controll
			motor_break();
			holdArm = FALSE;
			runcontrol = FALSE;
			// set our hold point
			holdTheta[0] = currentTheta[0];
			holdTheta[1] = currentTheta[1];
			// Clear everything
			for (i = 0; i<ERROR_DERIV_DELAY; i++) 
			{
				errorBuffer[0][i] = 0;
				errorBuffer[1][i] = 0;
			}
			errorIndex = 0;
			globalIndex = 0;
			controlCount = 0; 
			// read number of path points
			numPathPoints = read_short_int();
			// free previously allocated path memory
			free(pathTime);
			free(pathTheta1);
			free(pathTheta2);
			// allocate memory to store the path
			pathTime = (signed short int *) malloc(numPathPoints*sizeof(short int));			
			pathTheta1 = (signed short int *) malloc(numPathPoints*sizeof(short int));
			pathTheta2 = (signed short int *) malloc(numPathPoints*sizeof(short int));
			if ((pathTime == NULL) || (pathTheta1 == NULL) || (pathTheta2 == NULL)) {MEM_LED = 0;}  // <--- can't allocate memory (heap is probably too small)
			// read the path into the arrays
			for ( i = 0; i < numPathPoints; i++)
			{
				pathTime[i] = read_short_int();
				pathTheta1[i] = read_short_int();
				pathTheta2[i] = read_short_int();
				
			}
			// the last time point in the path is the period
			refPeriod = pathTime[numPathPoints - 1];
			// now that we know how long our period is, we can allocate the logging buffers
			// We log into these buffers:
			//refTheta1Log[recordIndex]
			//refTheta2Log[recordIndex]
			//currentTheta1Log[recordIndex]
			//currentTheta2Log[recordIndex]
			// And we record every RECORD_EVERY points
			// So we need refPeriod/RECORD_EVERY spots in the buffer to store the log
			// free the old mem first
			free(refTheta1Log);
			free(refTheta2Log);
			free(currentTheta1Log);
			free(currentTheta2Log);
			refTheta1Log = (signed short int *) malloc(refPeriod/RECORD_EVERY*sizeof(short int));
			refTheta2Log = (signed short int *) malloc(refPeriod/RECORD_EVERY*sizeof(short int));
			currentTheta1Log = (signed short int *) malloc(refPeriod/RECORD_EVERY*sizeof(short int));
			currentTheta2Log = (signed short int *) malloc(refPeriod/RECORD_EVERY*sizeof(short int));
			if ((refTheta1Log==NULL) || (refTheta2Log==NULL) || (currentTheta1Log==NULL) || (currentTheta2Log==NULL)) {MEM_LED = 0;}  // <--- can't allocate memory (heap is probably too small)
			// then hold
			holdArm = TRUE;
			runcontrol = FALSE;
			break;
			
		case C_SEND_LOG:
			// Send logging data
			sendLog=TRUE;
			break;
			
		case C_LOOP:
			// set the loop variable
			loop = read_short_int();
			break;

		case C_GOTO:
			// Set up a two point path from current position to the sent goto position, 
			// run it.
			// no loop
			// control params are gental
			// set the hold position to the final position
			
			// Stop
			motor_break();
			runcontrol = FALSE;
			holdArm = FALSE;
			// Clear everything
			for (i = 0; i<ERROR_DERIV_DELAY; i++) 
			{
				errorBuffer[0][i] = 0;
				errorBuffer[1][i] = 0;
			}
			errorIndex = 0;
			globalIndex = 0;
			controlCount = 0; 
			// Config and read path
			numPathPoints = 2;
			// free previously allocated path memory
			free(pathTime);
			free(pathTheta1);
			free(pathTheta2);
			// allocate memory to store the path
			pathTime = (signed short int *) malloc(numPathPoints*sizeof(short int));			
			pathTheta1 = (signed short int *) malloc(numPathPoints*sizeof(short int));
			pathTheta2 = (signed short int *) malloc(numPathPoints*sizeof(short int));
			if ((pathTime == NULL) || (pathTheta1 == NULL) || (pathTheta2 == NULL)) {MEM_LED = 0;}  // <--- can't allocate memory (heap is probably too small)
			// The first point of the path. Zero time and current position.
			pathTime[0] = 0;
			pathTheta1[0] = currentTheta[0];
			pathTheta2[0] = currentTheta[1];
			// The second point.
			pathTime[1] = GOTO_DURATION_TICKS;
			pathTheta1[1] = read_short_int();
			pathTheta2[1] = read_short_int();
			// the last time point in the path is the period
			refPeriod = pathTime[numPathPoints - 1];
			// now that we know how long our period is, we can allocate the logging buffers
			// We log into these buffers:
			//refTheta1Log[recordIndex]
			//refTheta2Log[recordIndex]
			//currentTheta1Log[recordIndex]
			//currentTheta2Log[recordIndex]
			// And we record every RECORD_EVERY points
			// So we need refPeriod/RECORD_EVERY spots in the buffer to store the log
			// free the old mem first
			free(refTheta1Log);
			free(refTheta2Log);
			free(currentTheta1Log);
			free(currentTheta2Log);
			refTheta1Log = (signed short int *) malloc(refPeriod/RECORD_EVERY*sizeof(short int));
			refTheta2Log = (signed short int *) malloc(refPeriod/RECORD_EVERY*sizeof(short int));
			currentTheta1Log = (signed short int *) malloc(refPeriod/RECORD_EVERY*sizeof(short int));
			currentTheta2Log = (signed short int *) malloc(refPeriod/RECORD_EVERY*sizeof(short int));
			if ((refTheta1Log==NULL) || (refTheta2Log==NULL) || (currentTheta1Log==NULL) || (currentTheta2Log==NULL)) {MEM_LED = 0;}  // <--- can't allocate memory (heap is probably too small)
			// Don't loop
			loop = FALSE;
			// set control params
			KPnum[0] = HOLD_KPnum;
			KPden[0] = HOLD_KPden;
			KDnum[0] = HOLD_KDnum;
			KDden[0] = HOLD_KDden;
			KInum[0] = HOLD_KInum;
			KIden[0] = HOLD_KIden;
			Kslope[0] = 60000;
			KPnum[1] = HOLD_KPnum;
			KPden[1] = HOLD_KPden;
			KDnum[1] = HOLD_KDnum;
			KDden[1] = HOLD_KDden;
			KInum[1] = HOLD_KInum;
			KIden[1] = HOLD_KIden;
			Kslope[1] = 60000;
			// So we'll hold at the end, set that point
			holdTheta[0] = pathTheta1[1];
			holdTheta[1] = pathTheta2[1];
			// And run it
			motor_run();
			break;
			
		case C_PAUSE:
			// hold the arm at current position
			holdArm = FALSE;
			runcontrol = FALSE;
			// set our hold point
			holdTheta[0] = currentTheta[0];
			holdTheta[1] = currentTheta[1];
			// clear errors
			for (i = 0; i<ERROR_DERIV_DELAY; i++) 
			{
				errorBuffer[0][i] = 0;
				errorBuffer[1][i] = 0;
			}
			errorIndex = 0;
			// then hold
			holdArm = TRUE;
			runcontrol = FALSE;
			break;
	
	}	// end of the giant switch statement

	// Toggle LED to indicate UART activity
	mLED_0_Toggle();

	return;

}

signed short int read_short_int()
{
	// reads two bytes into the union and returns it as a short int
	// (it should be easy to adapt this to 32 bit ints or floats, you just have to know the byte order)
	int j;
	union serialShortInt tempData;
	for (j=0; j<SHORT_INT_BYTELENGTH; j++){
					wait_timeout();
					tempData.dataChar[j]=ReadUART2();
	}
	return tempData.dataShortInt;
}

void wait_timeout()
{
	// waits until a byte is available.
	// the time out it just to prevent an infinite loop if the serial communication gets goofed up.
	unsigned int timeout=0;
	// 
	while (timeout<TIMEOUT_COUNT && !DataRdyUART2() ){
		timeout++;
	}
	
	if (timeout>=TIMEOUT_COUNT){
		//return -1;
		// This is bad, you should do some error handling here somehow!?
		mLED_1_On();
	}

}


void sendDataRS232()
{
	int i;
	char RS232_Out_Buffer[32]; // max characters per line (line feed and carriage return count)
	
	sprintf(RS232_Out_Buffer,"\n\rSTART\n\r"); //print START. MATLAB uses this as a start delimiter
	putsUART2(RS232_Out_Buffer);
	
	sprintf(RS232_Out_Buffer,"ROWS=%d\n\r", refPeriod/RECORD_EVERY); //print the number of rows, so matlab can make the correct sized buffer
	putsUART2(RS232_Out_Buffer);
	
	for(i = 0; i < refPeriod/RECORD_EVERY; i++)
	{
		LOOP_TIME_PIN = 0;
		sprintf(RS232_Out_Buffer,"%d %d %d %d\n\r", refTheta1Log[i], currentTheta1Log[i], refTheta2Log[i], currentTheta2Log[i]);	
		
		putsUART2(RS232_Out_Buffer);
		LOOP_TIME_PIN = 1;
	}
	
	sprintf(RS232_Out_Buffer,"END\n\r"); //output end so matlab knows we're done
	putsUART2(RS232_Out_Buffer);
	LOG_LED = 1; // led off
	
}



#endif
